/**********************************************************/
/* SAPS, version 1.0                                      */
/* Frank Hutter, Dave A.D. Tompkins & Holger H. Hoos      */
/* September 2002                                         */
/* ********************************************************/
/* This code is based on code in the WalkSAT framework    */
/* of Bram Cohen and Henry Kautz                          */
/**********************************************************/

/************************************/
/* Compilation flags                */
/************************************/
//=== If the constant NT is set the program compiles under Windows NT.
//=== Currently, the timing then reflects user time.
#ifdef WIN32
#define NT 1
#endif

/************************************/
/* Standard includes                */
/************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#ifdef NT
#include <time.h>
#else
#include <sys/time.h>
#include <sys/times.h>
#endif

#ifndef CLK_TCK
#define CLK_TCK 60
#endif

#ifdef NT
#define random() rand()
#define srandom(seed) srand(seed)
#endif

/***********************************************************/
/* Wrappers                                                */
/***********************************************************/
#define Lit(CLAUSE, POSITION) (clause[CLAUSE][POSITION])
#define Var(CLAUSE, POSITION) (ABS(Lit(CLAUSE,POSITION)))
static int scratch;
#define ABS(x) ((scratch=(x))>0?(scratch):(-scratch))

/**********************************************************/
/* Constants                                              */
/**********************************************************/
#define MAXATOM 100000			// maximum possible number of atoms
#define MAXCLAUSE 500000		// maximum possible number of clauses
#define MAXLENGTH 500       // max num of literals which can be in a clause
#define STOREBLOCK 2000000	// size of block to malloc each time

#define CWTYPE float        // type for clause weights
#define CWBIG 1E30f         // quasi-infinity for CWTYPE
#define BIG 100000000				// quasi-infinity for integer

/************************************/
/* Internal constants               */
/************************************/
#define TRUE 1
#define FALSE 0
#define NOVALUE -1

/**********************************************************/
/* Basic variables and data structures                    */
/**********************************************************/
int numAtoms;                   // number of atoms in input formula
int numClauses;                 // number of clauses in input formula
int numLiterals;                // number of literals in input formula
int try;                        // counter for independent tries
int flip;                       // counter for flips

//=== The numbering of atoms starts at 1. Not a is represented -a
//=== One dimensional arrays are statically allocated. Two dimensional 
//=== arrays are dynamically allocated in the second dimension only.
int atom[MAXATOM+1];				// value of each atom
int *clause[MAXCLAUSE];			// indexed as clause[clause_num][literal_num]
int size[MAXCLAUSE];				// length of each clause

int numFalse;									// number of false clauses
int falseClause[MAXCLAUSE];		// clauses which are false:
															// cFalse[i] = c <=> c is ith unsat. clause
int whereFalse[MAXCLAUSE];		// where each clause is listed in false:
															// whereFalse[c] = i <=> c is ith unsat. clause

int numTrueLit[MAXCLAUSE];			// number of true literals in each clause
int critVar[MAXCLAUSE];					// the literal for clauses critically sat
int *occurence[2*MAXATOM+1];		// where each literal occurs: indexed as
																// occurence[literal+MAXATOM][occurence_num]
int numOccurence[2*MAXATOM+1];	// number of times each literal occurs

int numCanMake;								// number of variables which can make a clause
int numMake[MAXATOM+1];				// number of clauses var makes when flipped.
int numBreak[MAXATOM+1];      // number of clauses var breaks when flipped.
int canMake[MAXATOM+1];       // variable can make clause (ie: numMake > 0)
int whereCanMake[MAXATOM+1];  // where variable appears in canMake[] list

int solution[MAXATOM+1];    // where the last solution gets saved
int targetCost = 0;         // cost to reach in order to have a solution
int cost[MAXCLAUSE];        // cost associated with each clause
int costAllFalse;           // sum of costs associated w/ all false clauses
int lowCost;                // best costAllFalse seen so far
int lowFlips;								// number of flips it took to reach lowCost

/**********************************************************/
/* DLS specific variables                                 */
/**********************************************************/
CWTYPE clauseWeight[MAXCLAUSE]; // weight of a clause
CWTYPE weightMake[MAXATOM+1];   // weight made by a variable
CWTYPE weightBreak[MAXATOM+1];  // weight broken by a variable
CWTYPE weightTotal;             // total clause weight
int nullFlip;                   // counter for nullFlips

/**********************************************************/
/* General parameters                                     */
/**********************************************************/
int seed;                       // seed used for randomization
int numTries = 10;              // number of tries
int cutoff = 100000;            // maximal number of flips per try
int printSol = FALSE;		        // flag determining whether to output solution

/**********************************************************/
/* SAPS parameters                                        */
/**********************************************************/
float smoothProb = 5.0f;				// default smoothing probability (in percent)
int noise = 1;                  // default noise (in percent)
float alpha = 1.3f;             // default scaling factor
float rho = 0.8f;               // default smoothing factor

int rsaps = FALSE;              // default: non-reactive
int lastImpr;                   // last improvement step, only used for RSAPS
int lastBest;                   // cost after last improvement step, only used for RSAPS

/**********************************************************/
/*  Global Statistics                                     */
/**********************************************************/
long int totalFlips = 0;					// total number of flips in all tries so far
long int totalSuccessFlips = 0;		// total num of flips in succesful tries so far
long flipsSinceLastSuccess = 0;		// flips since last satisfying assignment
double totalSuccessFlipsSqr = 0.0;// sum of squared success flips 
double meanSuccessFlips;					// avg. success flips
double stdDevSuccessFlips;				// standard deviation success flips
int totalNullFlips = 0;						// total num of nullFlips in all tries so far
int numSuccessTries = 0;					// total number of successful tries so far
int smoothStages;									// number of times the smoothing was performed

/**********************************************************/
/* Timing                                                 */
/**********************************************************/
double expertime;
#ifdef NT
#define random() rand()
#define srandom(seed) srand(seed)
clock_t ntStartTime;
#else
struct timeval tv;
struct timezone tzp;
#endif

/**********************************************************/
/*   Forward declarations                                 */
/**********************************************************/
void parseParameters(int argc,char *argv[]);
void scanInt(int argc, char *argv[], int i, int *varptr);
void scanFloat(int argc, char *argv[], int i, float *varptr);
void printSAPSHeader();
void printStatsHeader();
void initInstance();
void init();
int pickVar();
void flipAtom(int toflip);
void updateClauseWeights();
void smooth();
void updateStatsEndFlip();
void adaptSmoothing();
void updateStatsEndTry();
void printStatsEndTry();
CWTYPE computeCurrentCost();
double elapsedSeconds();
void printFinalStats();
void saveSolution();
void printSolution();

/************************************/
/* Main                             */
/************************************/
int main(int argc,char *argv[]){
  int var;
#ifdef NT
  ntStartTime = clock();
#else
  gettimeofday(&tv,&tzp);
  seed = (( tv.tv_sec & 0177 ) * 1000000) + tv.tv_usec;
#endif

  printSAPSHeader();
  
  parseParameters(argc, argv);
  srandom(seed);
  initInstance();
  printStatsHeader();
  elapsedSeconds();
  
  lowCost = BIG;
  for( try = 1; try <= numTries; try ++ ){
		init();
    flip = 0;
    while((costAllFalse > targetCost) && (flip < cutoff)) {
			flip++;
			var = pickVar();
			flipAtom(var);
			if (var==NOVALUE) updateClauseWeights();
			updateStatsEndFlip();
			if( rsaps ) adaptSmoothing();
    }
    updateStatsEndTry();
	  printStatsEndTry();
  }
  expertime = elapsedSeconds();
  printFinalStats();
  return 0;
}

int best[MAXATOM];
int pickVar(void){
	CWTYPE score, bestvalue;
  int i;
  register int numbest;	/* how many are tied for best */
  register int var;

  numbest = 0;
  bestvalue = CWBIG;
  for(i = 0;i < numCanMake;i++) {
		var = canMake[i];
		score = weightBreak[var] - weightMake[var];
		if (score<=bestvalue){
			if (score<bestvalue) numbest=0;
			bestvalue = score;
			best[numbest++] = var;
		}
  }

  if (bestvalue >= -1.0e-01f) {
		if (random()% 100 < noise) return ((random()%numAtoms)+1);
    else return NOVALUE;
  }
  return best[random()%numbest];
}

void updateClauseWeights(){
  int c,i,j,var,haveToDivide = FALSE;
  CWTYPE diff, old, divBorder = 1000;

  if (random()%1000 < smoothProb*10) smooth();
	//=== If one of the clause weights grows too much, divide all of them
	//=== by some constant. This does NOT affect the search heuristic
	//=== and is only done to avoid floating point errors.
  for(c = 0;c < numFalse;c++) {
	  if( clauseWeight[falseClause[c]] > divBorder ) {
		  haveToDivide = TRUE;
		  break;
	  }
  }

  if( haveToDivide ) {
	  weightTotal = 0.0;
    for(c = 0;c < numClauses;c++) {
	    old = clauseWeight[c];
      clauseWeight[c] /= divBorder;
      if (numTrueLit[c]==0) clauseWeight[c] *= alpha; // scaling.
      diff = clauseWeight[c] - old;
      if (numTrueLit[c]==0) {
	      for(j = 0;j < size[c];j++) weightMake[Var(c,j)] += diff;
      } else if (numTrueLit[c]==1) weightBreak[critVar[c]] += diff;
		  weightTotal += clauseWeight[c];
    }
  } else {
    for(i = 0;i < numFalse;i++) {
      c = falseClause[i];
      old = clauseWeight[c];
      clauseWeight[c] *= alpha; // scaling
      diff = clauseWeight[c] - old;
      for(j = 0;j < size[c];j++){
        var = ABS(clause[c][j]);
        weightMake[var] += diff;
      }
			weightTotal += diff;
    }
  }
}

/*
void smooth(){
// straight-forward implementation of smoothing step (as described in CP-02 paper)

  CWTYPE avg, diff, old;
  int j,c;
  avg = weightTotal / numClauses;
  avg *= (1.0f - rho);
  weightTotal = 0.0;
  for(c = 0;c < numClauses;c++) {
    old = clauseWeight[c];
    clauseWeight[c] *= rho;
    clauseWeight[c] += avg;
    diff = clauseWeight[c] - old;

    if (numTrueLit[c]==0) {
      for(j = 0;j < size[c];j++) weightMake[Var(c,j)] += diff;
    } else if (numTrueLit[c]==1) weightBreak[critVar[c]] += diff;
    weightTotal += clauseWeight[c];
  }
  smoothStages ++;
  if( rsaps ) smoothProb = 0; // because we just did it.
}
*/

void smooth() {
// slightly more efficient version (used for results in CP-02 paper)
// approximately 5% faster for uf200-* instances
  CWTYPE avg;
	int i,var;

  avg = weightTotal / numClauses;
  avg *= (1.0f - rho);

  for(i = 0;i < numClauses;i++) clauseWeight[i] += avg;
    
  weightTotal += (avg * numClauses);
  for(i = 1;i < numAtoms+1;i++) {
    var = i;
    weightMake[var] += numMake[var]*avg;
    weightBreak[var] += numBreak[var]*avg;
  }
  smoothStages++;
  if( rsaps) smoothProb = 0; // because we just did it.
}

void adaptSmoothing(){
	float theta = 0.16666f, delta = 0.1f;
    //=== Determine best value over last k steps.
	if ( flip-lastImpr > numClauses*theta ){
		smoothProb *= delta;	  
		lastImpr = flip;
		lastBest = numFalse;
	} else if (numFalse < lastBest) {
		smoothProb += (100-smoothProb) * 2 * delta;
    lastImpr=flip;
		lastBest = numFalse;
  }
}

void printSAPSHeader(){
  printf("+-------------------+\n");
  printf("| SAPS, version 1.0 +\n");
  printf("+-------------------+\n");
  printf("saps -h for help\n\n");
}

void printStatsHeader(){
  printf("Problem instance: numAtoms = %i, numClauses = %i, numLiterals = %i\n",numAtoms,numClauses,numLiterals);
  printf("General parameters: seed=%d, cutoff=%d, targetCost = %lf, print solution: ", seed, cutoff, targetCost );
  printf( printSol?"yes":"no");
  printf("\nSAPS parameters: noise=%d, alpha=%lf, rho=%lf, smoothProb", noise, alpha, rho);
  rsaps?printf(" is adapted reactively (RSAPS)"):printf(" = %f",smoothProb);
  printf("\ntries = %d\n\n\n\n",numTries);

  printf("  best    steps     final     total    scale   smooth            percent       avg      mean   std dev\n");
  printf("weight    until    #unsat     steps   stages   stages             2ndary    length     flips     flips\n");
  printf("  this     best      this     this      this     this  success     steps   success     until     until\n");
  printf("   try     soln       try      try       try      try     rate  this try     tries    assign    assign\n");
  fflush(stdout);
}

void printStatsEndTry(){
	printf("%6d %8d %9d %9ld %8ld %8ld", lowCost,lowFlips,numFalse,flip, nullFlip, smoothStages);
	printf("%9ld %9.2f ",(numSuccessTries*100)/try,(100*nullFlip+0.0)/flip);

  if (numSuccessTries > 0){
	  printf("%9d %9d", totalSuccessFlips/numSuccessTries, totalFlips/numSuccessTries);
    if (numSuccessTries > 1) printf(" %9.2f", stdDevSuccessFlips);
	}
	printf("\n");
  fflush(stdout);
}

void updateStatsEndTry(){
  totalFlips += flip;
  flipsSinceLastSuccess += flip;
    
  if(costAllFalse <= targetCost){
		saveSolution();
		numSuccessTries++;

		totalSuccessFlips += flip;
		totalSuccessFlipsSqr += ((double)flipsSinceLastSuccess)*((double)flipsSinceLastSuccess);
		meanSuccessFlips = (totalSuccessFlips+0.0) / numSuccessTries;
		if (numSuccessTries > 1){
		  stdDevSuccessFlips = sqrt( (numSuccessTries)/(numSuccessTries - 1.0) * ((totalSuccessFlipsSqr / numSuccessTries) -meanSuccessFlips*meanSuccessFlips));
		}	
		flipsSinceLastSuccess = 0;
	}
	if (computeCurrentCost() != costAllFalse) {
		fprintf(stderr, "Program error, verification of assignment cost fails!\n");
		exit(-1);
  }
}

void updateStatsEndFlip(){
  if (costAllFalse < lowCost) {
		lowCost = costAllFalse;
		lowFlips = flip;
		lastImpr = flip;
		lastBest = numFalse;
  }
}

void printFinalStats(void){
  printf("\ntotal elapsed seconds = %f\n", expertime);
  printf("average flips per second = %d\n", (long)(totalFlips/expertime));
	printf("proportion null flips = %f\n", ((double)totalNullFlips)/totalFlips);
  printf("number solutions found = %d\n", numSuccessTries);
  printf("final success rate = %f\n", ((double)numSuccessTries * 100.0)/numTries);
  printf("average length successful tries = %li\n", numSuccessTries ? (totalSuccessFlips/numSuccessTries) : 0);
  if (numSuccessTries > 0){
    printf("mean flips until assign = %f\n", meanSuccessFlips);
    if (numSuccessTries>1) printf("  standard deviation = %f\n", stdDevSuccessFlips);
  }

  if (numSuccessTries > 0){
		printf("ASSIGNMENT FOUND\n");
		if(printSol == TRUE) printSolution();
  } else printf("ASSIGNMENT NOT FOUND\n");
}

void init(){
  int i,j,var;
	weightTotal = 0;
  for(i = 0;i < numClauses;i++){
		numTrueLit[i] = 0;
		clauseWeight[i] = (CWTYPE) cost[i];
		weightTotal += clauseWeight[i];
	}

  numFalse = 0;
  costAllFalse = 0;
  nullFlip = 0;
  canMake[0] = 0;
	numCanMake = 0;
  smoothStages = 0;

  for(i = 1;i < numAtoms+1;i++){
		numBreak[i] = 0;
		numMake[i] = 0;
		canMake[i] = 0;
		weightBreak[i] = 0;
		weightMake[i] = 0;
  }

	for(i = 1;i < numAtoms+1;i++)
		atom[i] = random()%2;
    
  /* Initialize break and make in the following: */
  for(i = 0;i < numClauses;i++){
		for(j = 0;j < size[i];j++){
			if((Lit(i,j) > 0) == atom[Var(i,j)]){
				numTrueLit[i]++;
				critVar[i] = Var(i,j);
			}
		}
		if(numTrueLit[i] == 0){
			whereFalse[i] = numFalse;
			falseClause[numFalse] = i;
			numFalse++;
			costAllFalse += cost[i];
	
			for(j = 0;j < size[i];j++){
			// this variable makes clause i
				var = Var(i,j);
				numMake[var] += 1;
 				weightMake[var] += clauseWeight[i];
	
				if (numMake[var] == 1) {
					canMake[numCanMake] = var;
					whereCanMake[var] = numCanMake++;
				}
			}
		} else if (numTrueLit[i] == 1){
			// this variable breaks clause i
			numBreak[critVar[i]] += 1;
			weightBreak[critVar[i]] += clauseWeight[i];
		}
  }
	lastImpr = 0;
	lastBest = numFalse;
  lowCost = costAllFalse;
}

void flipAtom(int toflip){
  int i, j;					// loop counter
  int toEnforce;		// literal to enforce
  register int c,lit,var;
  int numocc;
  int *occptr;

  if (toflip == NOVALUE){
    totalNullFlips++;
    nullFlip++;
    return;
  }

  if(atom[toflip] > 0) toEnforce = -toflip;
  else toEnforce = toflip;

  //=== Flip the variable here.
  atom[toflip] = 1-atom[toflip];

  // Examine all occurrences of literals with toEnforce with old sign.
  // (these literals are made false; numTrueLit must be decremented in that clause)
  numocc = numOccurence[MAXATOM-toEnforce];
  occptr = occurence[MAXATOM-toEnforce];
  for(i = 0; i < numocc ;i++){
    // c = occurence[MAXATOM-toEnforce][i];
    // c is the number of the clause where the ith instance of the literal occurs
    c = *(occptr++);

    if (--numTrueLit[c] == 0){
      // Clause c is no longer satisfied by lit with toflip; it's now false.
      falseClause[numFalse] = c;
      whereFalse[c] = numFalse;
      numFalse++;
      costAllFalse += cost[c];

      //=== Decrement toflip's break.
			numBreak[toflip] --;
      weightBreak[toflip] -= clauseWeight[c];

      //=== Increment make of all vars in clause c.
      for (j=0; j<size[c]; j++){
				var = Var(c,j);
				if( numMake[var] == 0 ) {
					canMake[numCanMake] = var;
					whereCanMake[var] = numCanMake++;
				}
				numMake[var]++;
				weightMake[var] += clauseWeight[c];
      }
    } else if (numTrueLit[c] == 1){
      //=== Find the lit in this clause that makes it true, and inc its break.
      for (j=0; j<size[c]; j++){
        lit = Lit(c,j);
				var = ABS(lit);
				if((lit > 0) == atom[var]) {
					critVar[c] = var;
					numBreak[var] ++;
          weightBreak[var] += clauseWeight[c];
          break;
				}
      }
    }
  }

  // Examines all occurrences of literal with toEnforce with new sign */
  // (these literals are made true; numTrueLit must be incremented in that clause)
  numocc = numOccurence[MAXATOM+toEnforce];
  occptr = occurence[MAXATOM+toEnforce];
  for(i = 0; i < numocc; i++){
    // c = occurence[MAXATOM+toEnforce][i];
    // c is the number of the clause where the ith instance of the literal occurs
    c = *(occptr++);
    if (++numTrueLit[c] == 1){
      numFalse--;
      falseClause[whereFalse[c]] = falseClause[numFalse];
      whereFalse[falseClause[numFalse]] = whereFalse[c];
      costAllFalse -= cost[c];

      critVar[c] = toflip;
		  numBreak[toflip]++;
      weightBreak[toflip] += clauseWeight[c];

      //=== Decrement the weightMake of all vars in clause c.
      for (j=0; j<size[c]; j++){
				var = Var(c,j);
				numMake[var]--;
				weightMake[var] -= clauseWeight[c];

				if (numMake[var] == 0) {
					numCanMake--;
					canMake[whereCanMake[var]] = canMake[numCanMake];
					whereCanMake[canMake[numCanMake]] = whereCanMake[var];
				}
      }
    } else if (numTrueLit[c] == 2){
      //=== Find the lit in this clause other than toflip that makes it true,
      //=== and decrement its weightBreak.
      for (j=0; j<size[c]; j++){
				lit = Lit(c,j);
        var = ABS(lit);
				if( ((lit > 0) == atom[var]) && (toflip != var) ) {
					numBreak[var] --;
					weightBreak[var] -= clauseWeight[c];
					break;
				}
      }
    }
  }
}

void parseParameters(int argc,char *argv[]){
	int i;
	for (i=1;i < argc;i++){
		if (strcmp(argv[i],"-seed") == 0)							scanInt(argc,argv,++i,&seed);
		else if (strcmp(argv[i],"-cutoff") == 0)			scanInt(argc,argv,++i,&cutoff);
		else if (strcmp(argv[i],"-tries") == 0)				scanInt(argc,argv,++i,&numTries);
		else if (strcmp(argv[i],"-targetcost") == 0)	scanInt(argc,argv,++i,&targetCost);
		else if (strcmp(argv[i],"-noise") == 0)				scanInt(argc,argv,++i,&noise);
		else if (strcmp(argv[i],"-smoothprob") == 0)	scanFloat(argc,argv,++i,&smoothProb);
		else if (strcmp(argv[i],"-alpha") == 0)				scanFloat(argc,argv,++i,&alpha);
		else if (strcmp(argv[i],"-rho") == 0)					scanFloat(argc,argv,++i,&rho);
		else if (strcmp(argv[i],"-reactive") == 0)		rsaps = TRUE;
		else if (strcmp(argv[i],"-printsol") == 0)		printSol = TRUE;
		else {
			if (strcmp(argv[i],"-help")!=0 && strcmp(argv[i],"-h")!=0 )
				fprintf(stderr, "\nBad argument %s\n\n", argv[i]);
			fprintf(stderr, "General parameters:\n");
			fprintf(stderr, "  -seed N (integer, default: quasi-random based on system time)\n");
			fprintf(stderr, "  -cutoff N (integer, default: %d)\n", cutoff);
			fprintf(stderr, "  -tries N (integer, default: %d)\n", numTries);
			fprintf(stderr, "  -noise N (percent, default 1)\n", noise);
      fprintf(stderr, "  -alpha N (real number, default: 1.3)\n", alpha);
      fprintf(stderr, "  -rho N (real number, default: 0.8)\n", rho);
      fprintf(stderr, "  -smoothprob N (percent, default: 5)\n", smoothProb);
      fprintf(stderr, "  -reactive (RSAPS, adapts smoothProb)\n");
      fprintf(stderr, "  -targetcost N = find assignments of cost <= N (MAXSAT) (real number, default: 0)\n", targetCost);
      fprintf(stderr, "  -printsol\n");

			fprintf(stderr, "\nUsage:\n");
			fprintf(stderr, "  saps -parameters <myinstance.cnf\n\n");
      exit(-1);
    }
  }
}

void scanInt(int argc, char *argv[], int i, int *varptr){
  if (i>=argc || sscanf(argv[i],"%i",varptr)!=1){
    fprintf(stderr, "Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
    exit(-1);
  }
}

void scanFloat(int argc, char *argv[], int i, float *varptr){
  if (i>=argc || sscanf(argv[i],"%f",varptr)!=1){
    fprintf(stderr, "Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
    exit(-1);
  }
}

void initInstance(){
  int i,j,lit, freestore, lastc, nextc;
  int *storeptr;
  char buf[100];

  while ((lastc = getchar()) == 'c'){
    while ((nextc = getchar()) != EOF && nextc != '\n');
  }

  ungetc(lastc,stdin);
  if (scanf("p %s %i %i", buf,&numAtoms,&numClauses) != 3){
    fprintf(stderr,"Bad input file!\n");
    exit(-1);
  }
  if (strcmp("cnf", buf) != 0 ){
    fprintf(stderr, "Bad input file\n");
    exit(-1);
  }
  if(numAtoms > MAXATOM){
    fprintf(stderr,"ERROR - too many atoms\n");
    exit(-1);
  }

  freestore = 0;
  numLiterals = 0;
  for(i = 0;i < 2*MAXATOM+1;i++) numOccurence[i] = 0;

  for(i = 0;i < numClauses;i++){
    size[i] = -1;
    if (freestore < MAXLENGTH){
      storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
      freestore = STOREBLOCK;
    }
    clause[i] = storeptr;
    cost[i] = 1;

    do {
      size[i]++;
      if(size[i] > MAXLENGTH){
				fprintf(stderr, "ERROR - clause too long\n");
				exit(-1);
      }
      if (scanf("%i ",&lit) != 1){
				fprintf(stderr, "Bad input file\n");
				exit(-1);
      }
      if(lit != 0){
				*(storeptr++) = lit; // clause[i][size[i]] = j;
				freestore--;
				numLiterals++;
				numOccurence[lit+MAXATOM]++;
      }
    } while(lit != 0);
  }
  if(size[0] == 0){
    fprintf(stderr,"ERROR - incorrect problem format or extraneous characters\n");
    exit(-1);
  }

  for(i = 0;i < 2*MAXATOM+1;i++){
    if (freestore < numOccurence[i]){
      storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
      freestore = STOREBLOCK;
    }
    occurence[i] = storeptr;
    freestore -= numOccurence[i];
    storeptr += numOccurence[i];
    numOccurence[i] = 0;
  }

  for(i = 0;i < numClauses;i++){
    for(j = 0;j < size[i];j++){
      occurence[Lit(i,j)+MAXATOM][numOccurence[Lit(i,j)+MAXATOM]] = i;
      numOccurence[Lit(i,j)+MAXATOM]++;
    }
  }
}
CWTYPE computeCurrentCost(){
  int i, j, sat;
  CWTYPE result = 0;

  for (i=0;i < numClauses;i++){
		sat = FALSE;
		for (j=0; j < size[i]; j++){
	    if ( (Lit(i,j) > 0) == atom[Var(i,j)] ){
				sat = TRUE;
				break;
			}
		}
		if (!sat) result += cost[i];
	}
	return result;
}

#ifdef NT
double elapsedSeconds(void) {
	printf("\n\nNT !!!\n\n");
  return (((double)(clock() - ntStartTime)) / CLOCKS_PER_SEC);
}
#else
double elapsedSeconds(void){
  double answer;
  struct tms test1;
  times(&test1);
  answer = test1.tms_utime + test1.tms_cutime;
  return answer / 100;
}
#endif

void saveSolution(void){
  int i;
  for (i=1; i<=numAtoms; i++)	solution[i] = atom[i];
}

void printSolution(void){
  int i;
  for(i = 1;i < numAtoms+1;i++) printf("v %i\n", solution[i] == 1 ? i : -i);
}

